"use strict";

var _global = require("./global");

var _parser = _interopRequireDefault(require("./parser"));

var _validator = _interopRequireDefault(require("./validator"));

var _calculator = _interopRequireDefault(require("./calculator"));

var _reference = _interopRequireDefault(require("./reference"));

var _package = require("../package.json");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

module.exports.version = _package.version;
module.exports.knex = require('knex');

async function manipulate(knex, json, configs) {
  configs = configs || {};
  if (_global.JS._.isString(configs.serverName) && configs.serverName !== '') _global.JE.dbServer = configs.serverName;
  if (_global.JS._.isString(configs.primaryKey) && configs.primaryKey !== '') _global.JE.primaryKey = configs.primaryKey;
  if (_global.JS._.isString(configs.language) && configs.language !== '') _global.JE.i18nLang = configs.language;
  let lang = _global.JE.i18nLang;

  if (!_global.JS._.isUndefined(json.i18nLang)) {
    let langs = ['zh-cn', 'en-us'];
    if (langs.indexOf(json.i18nLang) >= 0) lang = json.i18nLang;
  }

  let returnObj = {
    code: 200,
    message: 'ok',
    data: {}
  };
  let logs = [];

  try {
    addLog(logs, [['zh-cn', '- 开启 JBDAP 任务'], ['en-us', '- JBDAP task begins']], lang);
    addLog(logs, [['zh-cn', '- 检查接收到的 JSON 是否合法'], ['en-us', '- Check JSON validity']], lang);

    _validator.default.checkJson(json, lang);

    addLog(logs, [['zh-cn', '* 识别用户身份'], ['en-us', '* Get user identity']], lang);
    let user = {};
    if (_global.JS._.isFunction(configs.recognizer)) user = await configs.recognizer(json.security || {}, lang);
    let root = {};
    let commands = json.commands;
    if (_global.JS._.isPlainObject(json.commands)) commands = [commands];
    addLog(logs, [['zh-cn', '- 开始处理接收到的指令'], ['en-us', '- Proceed to handle commands']], lang);
    await proceed(knex, configs.doorman, configs.scanner, configs.cacher, configs.dispatcher, commands, json.isTransaction, root, logs, user, lang);
    addLog(logs, [['zh-cn', '- 全部指令处理完成'], ['en-us', '- All commands handled successfully']], lang);
    addLog(logs, [['zh-cn', '- JBDAP 任务成功'], ['en-us', '- JBDAP task succeeded']], lang);

    for (let i = 0; i < commands.length; i++) {
      let item = commands[i];

      if (root[item.name].error) {
        returnObj.code = 500;
        returnObj.message = root[item.name].error;
      }

      if (item.return !== false) {
        returnObj.data[item.name] = root[item.name].data;
      }
    }

    if (json.needLogs === true) returnObj.logs = logs;
    return returnObj;
  } catch (err) {
    addLog(logs, [['zh-cn', '- JBDAP 任务失败'], ['en-us', '- JBDAP task failed']], lang);
    returnObj.code = 500;

    if (err.fullMessage) {
      let msg = err.fullMessage();

      if (msg.indexOf('DefError]') >= 0 || msg.indexOf('TypeError]') >= 0 || msg.indexOf('MissingError]') >= 0 || msg.indexOf('EmptyError]') >= 0 || msg.indexOf('ValueInvalidError]') >= 0 || msg.indexOf('SpilthError]') >= 0 || msg.indexOf('[SqlInjectionError]') >= 0) {
        returnObj.code = 400;
      }

      if (msg.indexOf('[AuthError]') >= 0) returnObj.code = 403;
      returnObj.message = err.fullMessage();
      returnObj.data = null;
      if (json.needLogs === true) returnObj.logs = logs;
      if (json.needTrace === true) returnObj.trace = err.fullStack();
    } else {
      returnObj.message = err.toString();
      returnObj.data = null;
      if (json.needLogs === true) returnObj.logs = logs;
      if (json.needTrace === true) returnObj.trace = err;
    }

    return returnObj;
  }
}

module.exports.manipulate = manipulate;

async function proceed(knex, doorman, scanner, cacher, dispatcher, commands, isTrans, root, logs, user, lang) {
  if (_global.JS._.isUndefined(lang)) lang = _global.JE.i18nLang;

  if (isTrans === true) {
    addLog(logs, [['zh-cn', '- 以事务方式执行'], ['en-us', '- Commands will be handled as a transaction']], lang);
    await knex.transaction(async function (trx) {
      try {
        for (let i = 0; i < commands.length; i++) {
          let cmd = commands[i];

          _validator.default.checkTopCommand(cmd, lang);

          addLog(logs, [['zh-cn', `${prefix(1)}$ 开始执行顶层命令 /${cmd.name} [${cmd.type} 类型]`], ['en-us', `${prefix(1)}$ Begin to handle top command /${cmd.name} ['${cmd.type}' type]`]], lang);
          await handleCmd(knex, trx, doorman, scanner, cacher, dispatcher, true, cmd, null, root, logs, user, commands, 1, lang);
          addLog(logs, [['zh-cn', `${prefix(1)}$ 顶层命令 /${cmd.name} 执行完毕`], ['en-us', `${prefix(1)}$ Top command /${cmd.name} finished`]], lang);
        }

        addLog(logs, [['zh-cn', '- 事务执行成功'], ['en-us', '- Transaction succeeded']], lang);
        return true;
      } catch (err) {
        addLog(logs, [['zh-cn', '- 事务失败，数据回滚'], ['en-us', '- Transaction failed, data has been rolled back']], lang);

        _global.JS.throwError('TransactionError', err, null, [['zh-cn', '解析或执行事务失败'], ['en-us', 'Proceed transaction failed']], lang);
      }
    });
  } else {
    try {
      addLog(logs, [['zh-cn', '- 非事务方式执行'], ['en-us', '- Commands will be handled in non-transaction mode']], lang);

      for (let i = 0; i < commands.length; i++) {
        let cmd = commands[i];

        _validator.default.checkTopCommand(cmd, lang);

        addLog(logs, [['zh-cn', `${prefix(1)}$ 开始执行顶层命令 /${cmd.name} [${cmd.type} 类型]`], ['en-us', `${prefix(1)}$ Begin to handle top command /${cmd.name} ['${cmd.type}' type]`]], lang);
        await handleCmd(knex, null, doorman, scanner, cacher, dispatcher, true, cmd, null, root, logs, user, commands, 1, lang);
        addLog(logs, [['zh-cn', `${prefix(1)}$ 顶层命令 /${cmd.name} 执行完毕`], ['en-us', `${prefix(1)}$ Top command /${cmd.name} finished`]], lang);
      }
    } catch (err) {
      _global.JS.throwError('CommandsExecError', err, null, [['zh-cn', '解析或执行指令失败'], ['en-us', 'Proceed commands failed']], lang);
    }
  }
}

async function handleCmd(knex, trx, doorman, scanner, cacher, dispatcher, isTop, cmd, parent, root, logs, user, commands, level, lang) {
  if (_global.JS._.isUndefined(lang)) lang = _global.JE.i18nLang;

  try {
    let result = null;

    if (isTop) {
      if (Object.keys(root).indexOf(cmd.name) >= 0) {
        addLog(logs, [['zh-cn', `${prefix(level)}$ /${cmd.name} 已经存在`], ['en-us', `${prefix(level)}$ /${cmd.name} already exists`]], lang);
        return root[cmd.name].data;
      }

      root[cmd.name] = {
        return: cmd.return || true,
        error: null,
        data: null
      };
    }

    let cr = await checkOnlyIf(cmd, parent, root, knex, trx, doorman, scanner, cacher, dispatcher, isTop, logs, user, commands, level, lang);

    if (cr === false) {
      addLog(logs, [['zh-cn', `${prefix(level)}* 前置条件不成立！跳过 '${cmd.name}' 指令`], ['en-us', `${prefix(level)}* Precondition does not match, skip '${cmd.name}' command`]], lang);
      return null;
    }

    if (cmd.target.indexOf('/') < 0) {
      if (cmd.target.indexOf('JBDAP_') === 0 || cmd.target.indexOf('SYS_') === 0) _global.JS.throwError('AuthError', null, null, [['zh-cn', `没有权限执行当前指令 '${cmd.name}'`], ['en-us', `No authority to handle current command '${cmd.name}'`]], lang);

      if (_global.JS._.isFunction(doorman)) {
        let data = null;
        let operations = ['update', 'delete', 'increase', 'decrease'];

        if (operations.indexOf(cmd.type) >= 0) {
          let qCmd = _global.JS._.cloneDeep(cmd);

          qCmd.type = 'list';
          qCmd.fields = ['*'];
          delete qCmd.data;
          data = await queryCmd(knex, trx, doorman, scanner, cacher, dispatcher, false, qCmd, parent, root, logs, user, commands, level, lang);
        }

        if (cmd.type === 'create') {
          if (_global.JS._.isPlainObject(cmd.data)) cmd.data = [cmd.data];
        }

        let authorized = true;

        try {
          let queryTypes = ['list', 'entity', 'values'];

          if (_global.JS._.isFunction(doorman)) {
            if (queryTypes.indexOf(cmd.type) < 0) authorized = await doorman(user, cmd, data, lang);
          }
        } catch (err) {
          _global.JS.throwError('AuthError', err, null, [['zh-cn', `不被许可执行当前指令 '${cmd.name}'`], ['en-us', `No permission to handle current command '${cmd.name}'`]], lang);
        }

        if (authorized === false) _global.JS.throwError('AuthError', null, null, [['zh-cn', `不被许可执行当前指令 '${cmd.name}'`], ['en-us', `No permission to handle current command '${cmd.name}'`]], lang);
      }
    }

    switch (cmd.type) {
      case 'entity':
      case 'list':
      case 'values':
        result = await queryCmd(knex, trx, doorman, scanner, cacher, dispatcher, isTop, cmd, parent, root, logs, user, commands, level, lang);
        break;

      case 'create':
      case 'update':
      case 'delete':
      case 'increase':
      case 'decrease':
        result = await executeCmd(knex, trx, doorman, scanner, cacher, dispatcher, isTop, cmd, parent, root, logs, user, commands, lang);
        break;

      case 'function':
        result = await executeFunction(knex, trx, doorman, scanner, cacher, dispatcher, isTop, cmd, parent, root, logs, user, commands, lang);
        break;
    }

    _validator.default.checkAfter(cmd.after, lang);

    if (!_global.JS._.isUndefined(cmd.after)) {
      addLog(logs, [['zh-cn', `${prefix(level + 1)}# 开始执行后置指令`], ['en-us', `${prefix(level + 1)}# Subsequent commands begin`]], lang);
      let afterCmds = cmd.after;
      if (!_global.JS._.isArray(cmd.after)) afterCmds = [cmd.after];

      for (let i = 0; i < afterCmds.length; i++) {
        let afterCmd = afterCmds[i];

        _validator.default.checkCommand(afterCmd, lang);

        addLog(logs, [['zh-cn', `${prefix(level + 2)}$ 开始执行指令${i + 1} /${afterCmd.name} [${afterCmd.type} 类型]`], ['en-us', `${prefix(level + 2)}$ Begin to handle command${i + 1} /${afterCmd.name} ['${afterCmd.type}' type]`]], lang);
        let temp = await handleCmd(knex, trx, doorman, scanner, cacher, dispatcher, false, afterCmd, result, root, logs, user, commands, level + 2, lang);
        addLog(logs, [['zh-cn', `${prefix(level + 2)}$ 指令${i + 1} /${afterCmd.name} 执行完毕]`], ['en-us', `${prefix(level + 2)}$ Command${i + 1} /${afterCmd.name} finished`]], lang);
      }

      addLog(logs, [['zh-cn', `${prefix(level + 1)}# 后置指令执行完毕`], ['en-us', `${prefix(level + 1)}# Subsequent commands finished`]], lang);
    }

    return result;
  } catch (err) {
    if (isTop === true) addLog(logs, [['zh-cn', `${prefix(level)}$ 执行顶层命令 /${cmd.name} 出错`], ['en-us', `${prefix(level)}$ Error occurred in top command /${cmd.name}`]], lang);else addLog(logs, [['zh-cn', `${prefix(level)}@ 执行级联指令 /${cmd.name} 出现错误`], ['en-us', `${prefix(level)}$ Error occurred in cascaded command /${cmd.name}`]], lang);

    _global.JS.throwError('CommandHandlerError', err, null, [['zh-cn', `处理指令 '${cmd.name}' 出错`], ['en-us', `Error occurred while handling command '${cmd.name}'`]], lang);
  }
}

function addLog(logs, dict, lang) {
  if (_global.JS._.isUndefined(lang)) lang = _global.JE.i18nLang;

  for (let i = 0; i < dict.length; i++) {
    let item = dict[i];

    if (item[0] === lang) {
      logs.push(item[1]);
      break;
    }
  }
}

function prefix(n) {
  let result = '';

  for (let i = 1; i < n; i++) result += '  ';

  return result;
}

async function queryCmd(knex, trx, doorman, scanner, cacher, dispatcher, isTop, cmd, parent, root, logs, user, commands, level, lang) {
  if (_global.JS._.isUndefined(lang)) lang = _global.JE.i18nLang;
  let result = null;

  try {
    let fields = _parser.default.parseFields(cmd.fields, lang);

    let rawFields = fields.raw;
    let cascadedFields = fields.cascaded;
    let valuesFields = fields.values;
    let toolingTypes = ['first', 'pick', 'clone'];
    let needList = _global.JS._.findIndex(valuesFields, item => {
      return toolingTypes.indexOf(item.operator) >= 0;
    }) >= 0;

    if (cmd.target.indexOf('/') === 0) {
      let objName = cmd.target.split('/')[1].split('.')[0];

      let idx = _global.JS._.findIndex(commands, {
        name: objName
      });

      if (Object.keys(root).indexOf(objName) < 0) {
        if (idx >= 0) {
          let command = commands[idx];

          _validator.default.checkTopCommand(command, lang);

          addLog(logs, [['zh-cn', `${prefix(level + 1)}$ 开始执行顶层命令 /${command.name} [${command.type} 类型]`], ['en-us', `${prefix(level + 1)}$ Begin to handle top command /${command.name} ['${command.type}' type]`]], lang);
          await handleCmd(knex, trx, doorman, scanner, cacher, dispatcher, isTop, command, parent, root, logs, user, commands, level + 1, lang);
          addLog(logs, [['zh-cn', `${prefix(level + 1)}$ 顶层命令 /${command.name} 执行完毕`], ['en-us', `${prefix(level + 1)}$ Top command /${command.name} finished`]], lang);
        } else _global.JS.throwError('RefDefError', null, null, [['zh-cn', `被引用对象 '/${objName}' 不存在于 commands 指令集中`], ['en-us', `The referred target '/${objName}' doesn't exist in commands`]], lang);
      }

      let rawData = root[objName].data;
      if (_global.JS._.isNull(rawData)) _global.JS.throwError('RefDefError', null, null, [['zh-cn', `被引用对象 '/${objName}' 不能为 null`], ['en-us', `The referred target '/${objName}' can not be a null`]], lang);

      if (_global.JS._.isPlainObject(rawData)) {
        let slices = cmd.target.split('/')[1].split('.');

        if (slices.length > 1) {
          for (let i = 1; i < slices.length; i++) {
            if (Object.keys(rawData).indexOf(slices[i]) < 0) _global.JS.throwError('RefDefError', null, null, [['zh-cn', `引用对象有误，不存在 '${cmd.target}' 路径`], ['en-us', `The referred path '${cmd.target}' does not exist`]], lang);
            rawData = rawData[slices[i]];
          }
        }
      }

      if (_global.JS._.isArray(rawData)) {
        if (cmd.type === 'entity') result = _reference.default.getObjFromList(rawData, cmd.query, rawFields, parent, root, lang);
        if (cmd.type === 'list') result = _reference.default.getListFromList(rawData, cmd.query, rawFields, parent, root, lang);
        if (cmd.type === 'values') result = _reference.default.getValuesFromList(rawData, valuesFields, lang);
      } else if (_global.JS._.isPlainObject(rawData)) {
        result = _reference.default.getObjFromObj(rawData, rawFields, lang);
      } else _global.JS.throwError('RefValueError', null, null, [['zh-cn', `被引用对象 '${cmd.target}' 必须是 Array 或者 Object 类型`], ['en-us', `The referred target '${cmd.target}' must be an Array or an Object`]], lang);
    } else {
      _validator.default.checkQuery(cmd.query, lang);

      let func = await getWhereFunc(cmd, parent, root, knex, trx, doorman, scanner, cacher, dispatcher, isTop, logs, user, commands, level, lang);

      let cmdCopy = _global.JS._.clone(cmd);

      cmdCopy.whereFunc = func;
      let hasCacher = false;
      let hasCache = false;

      if (_global.JS._.isFunction(cacher)) {
        hasCacher = true;
        let cache = cacher('get', cmdCopy);

        if (cache !== null) {
          result = cache;
          hasCache = true;
        }
      }

      if (!hasCache) {
        let query = knex.from(cmd.target);
        if (trx !== null) query = trx(cmd.target);

        if (cmd.type !== 'values' || cmd.type === 'values' && needList) {
          if (rawFields === '*') query = query.select();else query = query.column(rawFields).select();
        }

        if (!_global.JS._.isUndefined(cmd.query)) {
          if (func !== null) query = eval('query.where(' + func + ')');

          let order = _parser.default.parseOrder(cmd.query.order, lang);

          if (order.length > 0) query = query.orderBy(order);

          let pas = _parser.default.parseOffsetAndLimit(cmd.query.page, cmd.query.size, lang);

          if (pas.limit > 0) query = query.limit(pas.limit);
          if (pas.offset > 0) query = query.offset(pas.offset);
        }

        if (cmd.type === 'values' && needList === false) {
          if (valuesFields.length === 0) _global.JS.throwError('FieldsDefError', null, null, [['zh-cn', `'values' 查询类型至少要定义一个取值字段`], ['en-us', `Queries of type 'values' require at least one value field`]], lang);

          for (let i = 0; i < valuesFields.length; i++) {
            let item = valuesFields[i];
            let slices = item.fields.split(',');
            if (slices.length > 1) _global.JS.throwError('FieldsDefError', null, null, [['zh-cn', `'${item.operator}' 运算只接受一个字段`], ['en-us', `Calculations of type '${item.operator}' accept only one field`]], lang);
            query = eval(`query.${item.operator}({ ${item.name}: '${slices[0]}' })`);
          }

          console.log('sql:', query.toString());
          let list = await query;
          result = list[0];
        } else {
          console.log('sql:', query.toString());
          result = await query;
        }

        if (hasCacher) cacher('set', cmdCopy, result);
      }

      let authorized = true;

      try {
        if (_global.JS._.isFunction(doorman)) {
          authorized = await doorman(user, cmd, result, cmd.data, lang);
        }
      } catch (err) {
        _global.JS.throwError('AuthError', err, null, [['zh-cn', `不被许可执行当前指令 '${cmd.name}'`], ['en-us', `No permission to handle current command '${cmd.name}'`]], lang);
      }

      if (authorized === false) _global.JS.throwError('AuthError', null, null, [['zh-cn', `不被许可执行当前指令 '${cmd.name}'`], ['en-us', `No permission to handle current command '${cmd.name}'`]], lang);

      if (cmd.target.indexOf('/') < 0) {
        if (_global.JS._.isFunction(scanner)) result = await scanner(user, cmd, fields.raw, result, lang);
      }

      if (cmd.type === 'values' && needList === true) {
        if (valuesFields.length === 0) _global.JS.throwError('FieldsDefError', null, null, [['zh-cn', `'values' 查询类型至少要定义一个取值字段`], ['en-us', `Queries of type 'values' require at least one value field`]], lang);
        let values = {};

        for (let i = 0; i < valuesFields.length; i++) {
          let item = valuesFields[i];
          if (Object.keys(values).indexOf(item.name) >= 0) _global.JS.throwError('FieldsDefError', null, null, [['zh-cn', `'fields' 中定义的别名有重复`], ['en-us', `'fields' property contains conflict alias definition`]], lang);
          values[item.name] = _calculator.default.getValue(result, item, lang);
        }

        result = values;
      }
    }

    if (cmd.type !== 'values') {
      if (result.length > 0 && cascadedFields.length > 0) {
        for (let j = 0; j < cascadedFields.length; j++) {
          let command = cascadedFields[j];
          let key = command.name;

          _validator.default.checkCommand(command, lang);

          addLog(logs, [['zh-cn', `${prefix(level + 1)}$ 开始填充级联字段 [${key}]`], ['en-us', `${prefix(level + 1)}$ Begin to fill cascaded field [${fields}]`]], lang);

          for (let i = 0; i < result.length; i++) {
            if (cmd.type === 'entity' && i > 0) break;
            let item = result[i];
            item[key] = await handleCmd(knex, trx, doorman, scanner, cacher, dispatcher, false, command, item, root, logs, user, commands, level + 1, lang);
          }

          addLog(logs, [['zh-cn', `${prefix(level + 1)}$ 级联字段 [${key}] 填充完毕`], ['en-us', `${prefix(level + 1)}$ Filling cascaded field [${key}] finished`]], lang);
        }
      }
    }

    if (cmd.type === 'entity') {
      if (_global.JS._.isArray(result) && result.length === 0) result = null;else if (_global.JS._.isArray(result)) result = result[0];
      if (isTop) root[cmd.name].data = result;
    }

    if (cmd.type === 'list') {
      if (_global.JS._.isArray(result) && result.length === 0) result = null;
      if (isTop) root[cmd.name].data = result;
    }

    if (cmd.type === 'values') {
      if (isTop) root[cmd.name].data = result;
    }

    return result;
  } catch (err) {
    _global.JS.throwError('DBQueryError', err, null, [['zh-cn', `查询数据出错`], ['en-us', `Error occurred while querying data`]], lang);
  }
}

async function executeCmd(knex, trx, doorman, scanner, cacher, dispatcher, isTop, cmd, parent, root, logs, user, commands, lang) {
  if (_global.JS._.isUndefined(lang)) lang = _global.JE.i18nLang;
  let result = null;

  try {
    if (cmd.target.indexOf('/') >= 0) _global.JS.throwError('TargetDefError', null, null, [['zh-cn', `操作类指令 'target' 不能是引用对象`], ['en-us', `A referred object can not be the 'target' in '${cmd.type}' commands`]], lang);else {
      if (cmd.type !== 'create') _validator.default.checkQuery(cmd.query, lang);
      let query = knex(cmd.target);
      if (trx !== null) query = trx(cmd.target);

      if (!_global.JS._.isUndefined(cmd.query)) {
        let func = await getWhereFunc(cmd, parent, root, knex, trx, doorman, scanner, cacher, dispatcher, isTop, logs, user, commands, lang);
        if (func !== null) query = eval('query.where(' + func + ')');

        let order = _parser.default.parseOrder(cmd.query.order, lang);

        if (order.length > 0) query = query.orderBy(order);

        let pas = _parser.default.parseOffsetAndLimit(cmd.query.page, cmd.query.size, lang);

        if (pas.limit > 0) query = query.limit(pas.limit);
        if (pas.offset > 0) query = query.offset(pas.offset);
      }

      _validator.default.checkData(cmd.type, cmd.data, lang);

      if (cmd.type === 'delete') query = query.delete();else if (cmd.type === 'increase' || cmd.type === 'decrease') {
          if (_global.JS._.isString(cmd.data)) cmd.data = _parser.default.parseDataString(cmd.data, lang);
          if (cmd.type === 'increase') query = query.increment(cmd.data);
          if (cmd.type === 'decrease') query = query.decrement(cmd.data);
        } else {
          if (_global.JS._.isPlainObject(cmd.data)) cmd.data = [cmd.data];

          for (let i = 0; i < cmd.data.length; i++) {
            let item = cmd.data[i];
            let dbsvr = knex._context.client.config.client;

            if (cmd.type === 'create' && !item.createdAt) {
              if (dbsvr === 'mysql') item.createdAt = new Date().dtString();else if (dbsvr.indexOf('sqlite') >= 0) item.createdAt = new Date().toISOString();
            }

            if (!item.updatedAt) {
              if (dbsvr === 'mysql') item.updatedAt = new Date().dtString();else if (dbsvr.indexOf('sqlite') >= 0) item.updatedAt = new Date().toISOString();
            }

            let keys = Object.keys(item);

            for (let j = 0; j < keys.length; j++) {
              let key = keys[j];

              if (_global.JS._.isString(item[key])) {
                switch (item[key]) {
                  case 'JBDAP.fn.ISODate':
                    {
                      if (dbsvr === 'mysql') item[key] = new Date().dtString();else if (dbsvr.indexOf('sqlite') >= 0) item[key] = new Date().toISOString();
                      break;
                    }

                  default:
                    {
                      item[key] = _validator.default.safeString(item[key]);
                    }
                }
              }
            }
          }

          if (cmd.type === 'create') query = query.insert(cmd.data, [_global.JE.primaryKey]);
          if (cmd.type === 'update') query = query.update(cmd.data[0], [_global.JE.primaryKey]);
        }
      console.log('sql:', query.toString());
      result = await query;
      if (_global.JS._.isFunction(cacher)) cacher('clear', cmd);
      result = {
        dbServer: _global.JE.dbServer,
        return: result
      };
      if (isTop) root[cmd.name].data = result;
      return result;
    }
  } catch (err) {
    _global.JS.throwError('DBExecError', err, null, [['zh-cn', `操作数据出错`], ['en-us', `Error occurred while operating data`]], lang);
  }
}

async function getWhereFunc(cmd, parent, root, knex, trx, doorman, scanner, cacher, dispatcher, isTop, logs, user, commands, level, lang) {
  if (_global.JS._.isUndefined(lang)) lang = _global.JE.i18nLang;
  if (_global.JS._.isUndefined(cmd.query.where)) return null;
  if (Object.prototype.toString(cmd.query.where) === '[object Object]' && Object.keys(cmd.query.where).length === 0) return null;

  try {
    let where = {};
    if (Object.prototype.toString.call(cmd.query.where) === '[object Object]') where = cmd.query.where;
    let func = await getSubConditionFunc(where, 'and', cmd, parent, root, knex, trx, doorman, scanner, cacher, dispatcher, isTop, logs, user, commands, level, lang);
    return func;
  } catch (err) {
    _global.JS.throwError('WhereParserError', err, null, [['zh-cn', `'where' 条件解析出错`], ['en-us', `Error occurred while parsing 'where' conditions`]], lang);
  }
}

async function getSubConditionFunc(obj, type, cmd, parent, root, knex, trx, doorman, scanner, cacher, dispatcher, isTop, logs, user, commands, level, lang) {
  if (_global.JS._.isUndefined(lang)) lang = _global.JE.i18nLang;
  if (Object.prototype.toString.call(obj) !== '[object Object]') _global.JS.throwError('WhereDefError', null, null, [['zh-cn', `where 子查询条件不正确，'$${type}' 的值必须是 Object 类型`], ['en-us', `Invalid sub condition definition, value of '$${type}' must be an Object`]], lang);

  try {
    let keys = Object.keys(obj);
    let funcDefine = 'function(){ ';

    for (let i = 0; i < keys.length; i++) {
      let key = keys[i];
      let value = obj[key];

      if (key.indexOf('$') === 0) {
        let subType = key.split('$')[1];
        let subContent = await getSubConditionFunc(value, subType, cmd, parent, root, knex, trx, doorman, scanner, cacher, dispatcher, isTop, logs, user, commands, level, lang);
        let func = '';

        if (i === 0) {
          if (type === 'and' || type === 'or') func = `this.where({content})`;
          if (type === 'not') func = `this.whereNot({content})`;
        } else {
          if (type === 'and') func = `.andWhere({content})`;
          if (type === 'or') func = `.orWhere({content})`;
          if (type === 'not') func = `.whereNot({content})`;
        }

        let itemStr = func.replace('{content}', subContent);
        funcDefine += itemStr;
      } else {
          let comparision = _parser.default.parseComparision(key, value, lang);

          try {
            if (_global.JS._.isString(comparision.right)) {
              comparision.right = _calculator.default.tag2value(comparision.right, parent, root, null, lang);
              if (_validator.default.hasSqlInjection(comparision.right)) _global.JS.throwError('SqlInjectionError', null, null, [['zh-cn', `'where' 条件发现 sql 注入字符`], ['en-us', `Sql Injection characters found in 'where' conditions`]], lang);
            }
          } catch (err) {
            if (err.name === 'Tag2ValueError' && err.fullMessage().indexOf('[TagRefNotFilled]') > 0) {
              let ref = err.fullInfo().needRef;

              let idx = _global.JS._.findIndex(commands, {
                name: ref
              });

              if (idx >= 0) {
                let command = commands[idx];

                _validator.default.checkTopCommand(command, lang);

                addLog(logs, [['zh-cn', `${prefix(level + 1)}$ 开始执行顶层命令 /${command.name} [${command.type} 类型]`], ['en-us', `${prefix(level + 1)}$ Begin to handle top command /${command.name} ['${command.type}' type]`]], lang);
                await handleCmd(knex, trx, doorman, scanner, cacher, dispatcher, isTop, commands[idx], parent, root, logs, user, commands, level + 1, lang);
                addLog(logs, [['zh-cn', `${prefix(level + 1)}$ 顶层命令 /${command.name} 执行完毕`], ['en-us', `${prefix(level + 1)}$ Top command /${command.name} finished`]], lang);
                return await getSubConditionFunc(obj, type, cmd, parent, root, knex, trx, doorman, scanner, cacher, dispatcher, isTop, logs, user, commands, level, lang);
              } else _global.JS.throwError('WhereDefError', null, null, [['zh-cn', `'where' 查询中的被引用对象 '/${ref}' 不存在于 commands 指令集中`], ['en-us', `The referred target '/${ref}' in 'where' conditions doesn't exist in commands`]], lang);
            } else _global.JS.throwError('WhereValueError', err, null, [['zh-cn', `'where' 条件赋值出错`], ['en-us', `Error occurred while setting values for 'where' conditions`]], lang);
          }

          let left = comparision.left,
              right = comparision.right,
              operator = comparision.operator;

          switch (operator) {
            case 'eq':
              {
                let func = '';

                if (i === 0) {
                  if (type === 'and' || type === 'or') func = `this.where({'{left}':{right}})`;
                  if (type === 'not') func = `this.whereNot({'{left}':{right}})`;
                } else {
                  if (type === 'and') func = `.andWhere({'{left}':{right}})`;
                  if (type === 'or') func = `.orWhere({'{left}':{right}})`;
                  if (type === 'not') func = `.whereNot({'{left}':{right}})`;
                }

                let itemStr = func.replace('{left}', left).replace('{right}', JSON.stringify(right));
                funcDefine += itemStr;
                break;
              }

            case 'ne':
              {
                let func = '';

                if (i === 0) {
                  if (type === 'and' || type === 'or') func = `this.whereNot({'{left}':{right}})`;
                  if (type === 'not') func = `this.where({'{left}':{right}})`;
                } else {
                  if (type === 'and') func = `.andWhere(function(){this.whereNot('{left}',{right})})`;
                  if (type === 'or') func = `.orWhere(function(){this.whereNot('{left}',{right})})`;
                  if (type === 'not') func = `.where('{left}',{right})`;
                }

                let itemStr = func.replace('{left}', left).replace('{right}', JSON.stringify(right));
                funcDefine += itemStr;
                break;
              }

            case 'gte':
              {
                let func = '';

                if (i === 0) {
                  if (type === 'and' || type === 'or') func = `this.where('{left}','>=',{right})`;
                  if (type === 'not') func = `this.whereNot('{left}','>=',{right})`;
                } else {
                  if (type === 'and') func = `.andWhere('{left}','>=',{right})`;
                  if (type === 'or') func = `.orWhere('{left}','>=',{right})`;
                  if (type === 'not') func = `.whereNot('{left}','>=',{right})`;
                }

                let itemStr = func.replace('{left}', left).replace('{right}', JSON.stringify(right));
                funcDefine += itemStr;
                break;
              }

            case 'gt':
              {
                let func = '';

                if (i === 0) {
                  if (type === 'and' || type === 'or') func = `this.where('{left}','>',{right})`;
                  if (type === 'not') func = `this.whereNot('{left}','>',{right})`;
                } else {
                  if (type === 'and') func = `.andWhere('{left}','>',{right})`;
                  if (type === 'or') func = `.orWhere('{left}','>',{right})`;
                  if (type === 'not') func = `.whereNot('{left}','>',{right})`;
                }

                let itemStr = func.replace('{left}', left).replace('{right}', JSON.stringify(right));
                funcDefine += itemStr;
                break;
              }

            case 'lte':
              {
                let func = '';

                if (i === 0) {
                  if (type === 'and' || type === 'or') func = `this.where('{left}','<=',{right})`;
                  if (type === 'not') func = `this.whereNot('{left}','<=',{right})`;
                } else {
                  if (type === 'and') func = `.andWhere('{left}','<=',{right})`;
                  if (type === 'or') func = `.orWhere('{left}','<=',{right})`;
                  if (type === 'not') func = `.whereNot('{left}','<=',{right})`;
                }

                let itemStr = func.replace('{left}', left).replace('{right}', JSON.stringify(right));
                funcDefine += itemStr;
                break;
              }

            case 'lt':
              {
                let func = '';

                if (i === 0) {
                  if (type === 'and' || type === 'or') func = `this.where('{left}','<',{right})`;
                  if (type === 'not') func = `this.whereNot('{left}','<',{right})`;
                } else {
                  if (type === 'and') func = `.andWhere('{left}','<',{right})`;
                  if (type === 'or') func = `.orWhere('{left}','<',{right})`;
                  if (type === 'not') func = `.whereNot('{left}','<',{right})`;
                }

                let itemStr = func.replace('{left}', left).replace('{right}', JSON.stringify(right));
                funcDefine += itemStr;
                break;
              }

            case 'in':
              {
                let func = '';

                if (i === 0) {
                  if (type === 'and' || type === 'or') func = `this.whereIn('{left}',{right})`;else func = `this.whereNotIn('{left}',{right})`;
                } else {
                  if (type === 'and') func = `.andWhere(function(){this.whereIn('{left}',{right})})`;
                  if (type === 'or') func = `.orWhereIn('{left}',{right})`;
                  if (type === 'not') func = `.whereNotIn('{left}',{right})`;
                }

                let itemStr = func.replace('{left}', left).replace('{right}', JSON.stringify(right));
                funcDefine += itemStr;
                break;
              }

            case 'notIn':
              {
                let func = '';

                if (i === 0) {
                  if (type === 'and' || type === 'or') func = `this.whereNotIn('{left}',{right})`;else func = `this.whereIn('{left}',{right})`;
                } else {
                  if (type === 'and') func = `.andWhere(function(){this.whereNotIn('{left}',{right})})`;
                  if (type === 'or') func = `.orWhereNotIn('{left}',{right})`;
                  if (type === 'not') func = `.whereIn('{left}',{right})`;
                }

                let itemStr = func.replace('{left}', left).replace('{right}', JSON.stringify(right));
                funcDefine += itemStr;
                break;
              }

            case 'like':
              {
                let func = '';

                if (i === 0) {
                  if (type === 'and' || type === 'or') func = `this.where('{left}','like',{right})`;
                  if (type === 'not') func = `this.whereNot('{left}','like',{right})`;
                } else {
                  if (type === 'and') func = `.andWhere('{left}','like',{right})`;
                  if (type === 'or') func = `.orWhere('{left}','like',{right})`;
                  if (type === 'not') func = `.whereNot('{left}','like',{right})`;
                }

                let itemStr = func.replace('{left}', left).replace('{right}', JSON.stringify(right));
                funcDefine += itemStr;
                break;
              }

            case 'notLike':
              {
                let func = '';

                if (i === 0) {
                  if (type === 'and' || type === 'or') func = `this.whereNot('{left}','like',{right})`;
                  if (type === 'not') func = `this.where('{left}','like',{right})`;
                } else {
                  if (type === 'and') func = `.andWhere(function(){this.whereNot('{left}','like',{right})})`;
                  if (type === 'or') func = `.orWhereNot('{left}','like',{right})`;
                  if (type === 'not') func = `.where('{left}','like',{right})`;
                }

                let itemStr = func.replace('{left}', left).replace('{right}', JSON.stringify(right));
                funcDefine += itemStr;
                break;
              }

            case 'contains':
              {
                let func = '';

                if (i === 0) {
                  if (type === 'and' || type === 'or') func = `this.where('{left}','like',${JSON.stringify('%' + right + '%')})`;
                  if (type === 'not') func = `this.whereNot('{left}','like',${JSON.stringify('%' + right + '%')})`;
                } else {
                  if (type === 'and') func = `.andWhere('{left}','like',${JSON.stringify('%' + right + '%')})`;
                  if (type === 'or') func = `.orWhere('{left}','like',${JSON.stringify('%' + right + '%')})`;
                  if (type === 'not') func = `.whereNot('{left}','like',${JSON.stringify('%' + right + '%')})`;
                }

                let itemStr = func.replace('{left}', left);
                funcDefine += itemStr;
                break;
              }

            case 'notContain':
              {
                let func = '';

                if (i === 0) {
                  if (type === 'and' || type === 'or') func = `this.whereNot('{left}','like',${JSON.stringify('%' + right + '%')})`;
                  if (type === 'not') func = `this.where('{left}','like',${JSON.stringify('%' + right + '%')})`;
                } else {
                  if (type === 'and') func = `.andWhere(function(){this.whereNot('{left}','like',${JSON.stringify('%' + right + '%')})})`;
                  if (type === 'or') func = `.orWhereNot('{left}','like',${JSON.stringify('%' + right + '%')})`;
                  if (type === 'not') func = `.where('{left}','like',${JSON.stringify('%' + right + '%')})`;
                }

                let itemStr = func.replace('{left}', left);
                funcDefine += itemStr;
                break;
              }

            case 'startsWith':
              {
                let func = '';

                if (i === 0) {
                  if (type === 'and' || type === 'or') func = `this.where('{left}','like',${JSON.stringify(right + '%')})`;
                  if (type === 'not') func = `this.whereNot('{left}','like',${JSON.stringify(right + '%')})`;
                } else {
                  if (type === 'and') func = `.andWhere('{left}','like',${JSON.stringify(right + '%')})`;
                  if (type === 'or') func = `.orWhere('{left}','like',${JSON.stringify(right + '%')})`;
                  if (type === 'not') func = `.whereNot('{left}','like',${JSON.stringify(right + '%')})`;
                }

                let itemStr = func.replace('{left}', left);
                funcDefine += itemStr;
                break;
              }

            case 'notStartWith':
              {
                let func = '';

                if (i === 0) {
                  if (type === 'and' || type === 'or') func = `this.whereNot('{left}','like',${JSON.stringify(right + '%')})`;
                  if (type === 'not') func = `this.where('{left}','like',${JSON.stringify(right + '%')})`;
                } else {
                  if (type === 'and') func = `.andWhere(function(){this.whereNot('{left}','like',${JSON.stringify(right + '%')})})`;
                  if (type === 'or') func = `.orWhereNot('{left}','like',${JSON.stringify(right + '%')})`;
                  if (type === 'not') func = `.where('{left}','like',${JSON.stringify(right + '%')})`;
                }

                let itemStr = func.replace('{left}', left);
                funcDefine += itemStr;
                break;
              }

            case 'endsWith':
              {
                let func = '';

                if (i === 0) {
                  if (type === 'and' || type === 'or') func = `this.where('{left}','like',${JSON.stringify('%' + right)})`;
                  if (type === 'not') func = `this.whereNot('{left}','like',${JSON.stringify('%' + right)})`;
                } else {
                  if (type === 'and') func = `.andWhere('{left}','like',${JSON.stringify('%' + right)})`;
                  if (type === 'or') func = `.orWhere('{left}','like',${JSON.stringify('%' + right)})`;
                  if (type === 'not') func = `.whereNot('{left}','like',${JSON.stringify('%' + right)})`;
                }

                let itemStr = func.replace('{left}', left);
                funcDefine += itemStr;
                break;
              }

            case 'notEndWith':
              {
                let func = '';

                if (i === 0) {
                  if (type === 'and' || type === 'or') func = `this.whereNot('{left}','like',${JSON.stringify('%' + right)})`;
                  if (type === 'not') func = `this.where('{left}','like',${JSON.stringify('%' + right)})`;
                } else {
                  if (type === 'and') func = `.andWhere(function(){this.whereNot('{left}','like',${JSON.stringify('%' + right)})})`;
                  if (type === 'or') func = `.orWhereNot('{left}','like',${JSON.stringify('%' + right)})`;
                  if (type === 'not') func = `.where('{left}','like',${JSON.stringify('%' + right)})`;
                }

                let itemStr = func.replace('{left}', left);
                funcDefine += itemStr;
                break;
              }

            case 'between':
              {
                let func = '';

                if (i === 0) {
                  if (type === 'and' || type === 'or') func = `this.whereBetween('{left}',{right})`;else func = `this.whereNotBetween('{left}',{right})`;
                } else {
                  if (type === 'and') func = `.andWhere(function(){this.whereBetween('{left}',{right})})`;
                  if (type === 'or') func = `.orWhereBetween('{left}',{right})`;
                  if (type === 'not') func = `.whereNotBetween('{left}',{right})`;
                }

                let itemStr = func.replace('{left}', left).replace('{right}', JSON.stringify(right));
                funcDefine += itemStr;
                break;
              }

            case 'notBetween':
              {
                let func = '';

                if (i === 0) {
                  if (type === 'and' || type === 'or') func = `this.whereNotBetween('{left}',{right})`;else func = `this.whereBetween('{left}',{right})`;
                } else {
                  if (type === 'and') func = `.andWhere(function(){this.whereNotBetween('{left}',{right})})`;
                  if (type === 'or') func = `.orWhereNotBetween('{left}',{right})`;
                  if (type === 'not') func = `.whereBetween('{left}',{right})`;
                }

                let itemStr = func.replace('{left}', left).replace('{right}', JSON.stringify(right));
                funcDefine += itemStr;
                break;
              }

            case 'isNull':
              {
                let func = '';

                if (i === 0) {
                  if (type === 'and' || type === 'or') func = `this.whereNull('{left}')`;else func = `this.whereNotNull('{left}')`;
                } else {
                  if (type === 'and') func = `.andWhere(function(){this.whereNull('{left}')})`;
                  if (type === 'or') func = `.orWhereNull('{left}')`;
                  if (type === 'not') func = `.whereNotNull('{left}')`;
                }

                let itemStr = func.replace('{left}', left);
                funcDefine += itemStr;
                break;
              }

            case 'isNotNull':
              {
                let func = '';

                if (i === 0) {
                  if (type === 'and' || type === 'or') func = `this.whereNotNull('{left}')`;else func = `this.whereNull('{left}')`;
                } else {
                  if (type === 'and') func = `.andWhere(function(){this.whereNotNull('{left}')})`;
                  if (type === 'or') func = `.orWhereNotNull('{left}')`;
                  if (type === 'not') func = `.whereNull('{left}')`;
                }

                let itemStr = func.replace('{left}', left);
                funcDefine += itemStr;
                break;
              }

            default:
              _global.JS.throwError('WhereDefError', null, null, [['zh-cn', `运算符 '${operator}' 不存在`], ['en-us', `Operator '${operator}' does not exist`]], lang);

          }
        }
    }

    funcDefine += ' }';
    return funcDefine;
  } catch (err) {
    _global.JS.throwError('SubWhereParserError', err, null, [['zh-cn', `'where' 子条件解析出错`], ['en-us', `Error occurred while parsing 'where' sub conditions`]], lang);
  }
}

async function executeFunction(knex, trx, doorman, scanner, cacher, dispatcher, isTop, cmd, parent, root, logs, user, commands, lang) {
  if (_global.JS._.isUndefined(lang)) lang = _global.JE.i18nLang;

  try {
    let result = null;
    if (cmd.target.indexOf('/') >= 0) _global.JS.throwError('TargetDefError', null, null, [['zh-cn', `服务端函数指令 'target' 不能是引用对象`], ['en-us', `A referred object can not be the 'target' in server-side function commands`]], lang);else {
      if (_global.JS._.isFunction(dispatcher)) {
        let data = cmd.data;
        if (!_global.JS._.isPlainObject(data) && !_global.JS._.isString(data)) _global.JS.throwError('PropTypeError', null, null, [['zh-cn', `对于 'function' 操作，属性 'data' 必须是 String|Object 类型`], ['en-us', `Property 'data' must be a String or an Object for 'function' type operations`]], lang);

        try {
          if (_global.JS._.isString(data) && data !== '') {
            cmd.data = _calculator.default.tag2value(data, parent, root, null, lang);
            if (_validator.default.hasSqlInjection(cmd.data)) _global.JS.throwError('SqlInjectionError', null, null, [['zh-cn', `发现 sql 注入字符`], ['en-us', `Sql Injection characters found`]], lang);
          }

          if (_global.JS._.isPlainObject(data) && Object.keys(data).length > 0) {
            let keys = Object.keys(data);

            for (let i = 0; i < keys.length; i++) {
              let prop = data[keys[i]];

              if (_global.JS._.isString(prop)) {
                cmd.data[keys[i]] = _calculator.default.tag2value(prop, parent, root, null, lang);
                if (_validator.default.hasSqlInjection(cmd.data[keys[i]])) _global.JS.throwError('SqlInjectionError', null, null, [['zh-cn', `发现 sql 注入字符`], ['en-us', `Sql Injection characters found`]], lang);
              }
            }
          }

          result = await dispatcher(user, knex, trx, cmd.target, cmd.data, lang);
          if (isTop) root[cmd.name].data = result;
          return result;
        } catch (err) {
          if (err.name === 'Tag2ValueError' && err.fullMessage().indexOf('[TagRefNotFilled]') > 0) {
            let ref = err.fullInfo().needRef;

            let idx = _global.JS._.findIndex(commands, {
              name: ref
            });

            if (idx >= 0) {
              let command = commands[idx];

              _validator.default.checkTopCommand(command, lang);

              addLog(logs, [['zh-cn', `${prefix(level + 1)}$ 开始执行顶层命令 /${command.name} [${command.type} 类型]`], ['en-us', `${prefix(level + 1)}$ Begin to handle top command /${command.name} ['${command.type}' type]`]], lang);
              await handleCmd(knex, trx, doorman, scanner, cacher, dispatcher, isTop, commands[idx], parent, root, logs, user, commands, level + 1, lang);
              addLog(logs, [['zh-cn', `${prefix(level + 1)}$ 顶层命令 /${command.name} 执行完毕`], ['en-us', `${prefix(level + 1)}$ Top command /${command.name} finished`]], lang);
              return await executeFunction(knex, trx, doorman, scanner, cacher, dispatcher, isTop, cmd, parent, root, logs, user, commands, lang);
            } else _global.JS.throwError('DefError', null, null, [['zh-cn', `被引用对象 '/${ref}' 不存在于 commands 指令集中`], ['en-us', `The referred target '/${ref}' doesn't exist in commands`]], lang);
          } else _global.JS.throwError('ValueError', err, null, [['zh-cn', `'data' 赋值出错`], ['en-us', `Error occurred while setting values for 'data'`]], lang);
        }
      } else _global.JS.throwError('JBDAPConfigError', null, null, [['zh-cn', `服务器没有配置 dispatcher 调度器`], ['en-us', `No 'dispatcher' was set to server, but some server-side functions were called`]], lang);
    }
  } catch (err) {
    _global.JS.throwError('FunctionError', err, null, [['zh-cn', `执行服务端函数 '${cmd.name}' 出错`], ['en-us', `Error occurred while executing server-side function '${cmd.name}'`]], lang);
  }
}

async function checkOnlyIf(cmd, parent, root, knex, trx, doorman, scanner, cacher, dispatcher, isTop, logs, user, commands, level, lang) {
  if (_global.JS._.isUndefined(lang)) lang = _global.JE.i18nLang;

  try {
    if (_global.JS._.isUndefined(cmd.onlyIf)) return true;

    if (Object.prototype.toString.call(cmd.onlyIf) === '[object Object]') {
      return _calculator.default.checkCondition(cmd.onlyIf, 'and', parent, root, null, lang);
    }
  } catch (err) {
    if (err.name === 'ConditionCalError' && err.fullMessage().indexOf('[TagRefNotFilled]') > 0) {
      let ref = err.fullInfo().needRef;

      let idx = _global.JS._.findIndex(commands, {
        name: ref
      });

      if (idx >= 0) {
        let command = commands[idx];

        _validator.default.checkTopCommand(command, lang);

        addLog(logs, [['zh-cn', `${prefix(level + 1)}$ 开始执行顶层命令 /${command.name} [${command.type} 类型]`], ['en-us', `${prefix(level + 1)}$ Begin to handle top command /${command.name} ['${command.type}' type]`]], lang);
        await handleCmd(knex, trx, doorman, scanner, cacher, dispatcher, isTop, commands[idx], parent, root, logs, user, commands, level + 1, lang);
        addLog(logs, [['zh-cn', `${prefix(level + 1)}$ 顶层命令 /${command.name} 执行完毕`], ['en-us', `${prefix(level + 1)}$ Top command /${command.name} finished`]], lang);
        return await checkOnlyIf(cmd, parent, root, knex, trx, doorman, scanner, cacher, dispatcher, isTop, logs, user, commands, level, lang);
      } else _global.JS.throwError('OnlyIfDefError', null, null, [['zh-cn', `'onlyIf' 条件中的被引用对象 '/${ref}' 不存在于 commands 指令集中`], ['en-us', `The referred target '/${ref}' in 'onlyIf' conditions doesn't exist in commands`]], lang);
    } else _global.JS.throwError('OnlyIfError', err, null, [['zh-cn', `'onlyIf' 条件解析出错`], ['en-us', `Error occurred while parsing 'onlyIf' conditions`]], lang);
  }
}